use super::{BundlePoller, Error, PollResult, Result};
use crate::bundle::Bundle;
use async_compat::CompatExt;
use async_tar::Archive;
use async_trait::async_trait;
use rusoto_core::RusotoError;
use rusoto_s3::{
  GetObjectError, GetObjectRequest, HeadObjectError, HeadObjectRequest, S3Client, S3,
};
use snafu::{ResultExt, Snafu};
use std::path::PathBuf;

#[derive(Snafu, Debug)]
pub enum InternalError {
  S3HeadObjectError {
    source: RusotoError<HeadObjectError>,
  },
  S3GetObjectError {
    source: RusotoError<GetObjectError>,
  },
  S3MissingBody,
  S3MissingETag,
}

impl From<InternalError> for Error {
  fn from(err: InternalError) -> Self {
    Error::InternalPollerError {
      source: Box::new(err),
    }
  }
}

pub struct S3BundlePoller {
  client: S3Client,
  bucket: String,
  object_name: String,
}

impl S3BundlePoller {
  pub fn new(
    access_key: String,
    secret_key: String,
    endpoint: String,
    bucket: String,
    region: String,
    object_name: String,
  ) -> S3BundlePoller {
    let region = rusoto_core::Region::Custom {
      name: region,
      endpoint,
    };
    let credentials = rusoto_credential::StaticProvider::new_minimal(access_key, secret_key);
    let dispatcher = rusoto_core::HttpClient::new().unwrap();
    let client = S3Client::new_with(dispatcher, credentials, region);

    S3BundlePoller {
      client,
      bucket,
      object_name,
    }
  }
}

#[async_trait]
impl BundlePoller for S3BundlePoller {
  async fn poll(&self, active_bundle: &Option<Bundle>) -> Result<PollResult> {
    if let Some(active_bundle) = active_bundle {
      if active_bundle.etag.is_some() {
        let mut head_request: HeadObjectRequest = Default::default();

        head_request.bucket = self.bucket.clone();
        head_request.key = self.object_name.clone();

        let head_response = self
          .client
          .head_object(head_request)
          .await
          .context(S3HeadObjectError)?;

        if head_response.e_tag == active_bundle.etag {
          info!(
            "S3BundlePoller: No updates found for object: {}",
            self.object_name
          );

          // Skip update download
          return Ok(PollResult::Skip);
        } else {
          info!(
            "S3BundlePoller: Object {} has changed. Starting bundle update.",
            self.object_name
          );
        }
      } else {
        error!("S3BundlePoller: Expected the active bundle to have a valid ETag.");

        return Err(Error::MissingBundleID);
      }
    }

    // Head object
    let mut head_request: HeadObjectRequest = Default::default();

    head_request.bucket = self.bucket.clone();
    head_request.key = self.object_name.clone();

    let head_response = self
      .client
      .head_object(head_request)
      .await
      .context(S3HeadObjectError)?;

    Ok(PollResult::UpdateReady {
      etag: head_response.e_tag.ok_or(InternalError::S3MissingETag)?,
    })
  }

  async fn retrieve(&self, bundle_id: &str, path: PathBuf) -> Result<()> {
    info!("S3BundlePoller: Starting bundle download...");

    let mut get_object_request: GetObjectRequest = Default::default();

    get_object_request.bucket = self.bucket.clone();
    get_object_request.key = self.object_name.clone();

    let get_object_response = self
      .client
      .get_object(get_object_request)
      .await
      .context(S3GetObjectError)?;

    let current_bundle_id = get_object_response
      .e_tag
      .ok_or(InternalError::S3MissingETag)?;

    if current_bundle_id != bundle_id {
      return Err(Error::MismatchedBundleID {
        original_id: String::from(bundle_id),
        current_id: current_bundle_id,
      });
    }

    info!("S3BundlePoller: Unpacking bundle...");

    // TODO: Write temp file to disk.
    let archive = Archive::new(
      get_object_response
        .body
        .ok_or(InternalError::S3MissingBody)?
        .into_async_read()
        .compat(),
    );

    archive.unpack(path).await.context(super::UnpackError)?;

    Ok(())
  }
}
