use crate::thread::{self, ThreadHandle};
use crate::{
  bundle::{Unbundler, UnbundlerStatus},
  config::StatsConfig,
};
use actix::System;
use actix_web::{dev::Server, middleware::Logger, App, HttpResponse, HttpServer, Responder};
use mpsc::{RecvError, SendError, Sender};
use serde::Serialize;
use snafu::{ResultExt, Snafu};
use std::{
  path::PathBuf,
  sync::{mpsc, Arc},
  thread::{JoinHandle, Thread},
};

#[derive(Debug, Snafu)]
pub enum Error {
  ChannelSend { source: SendError<Server> },
  ChannelReceive { source: RecvError },
  Bind { source: std::io::Error },
  SystemRun { source: std::io::Error },
}

pub type Result<T, E = Error> = std::result::Result<T, E>;

struct State {
  unbundler: Arc<Unbundler>,
}

pub struct StatsServer {
  config: StatsConfig,
  unbundler: Arc<Unbundler>,
}

impl StatsServer {
  pub fn new(config: StatsConfig, unbundler: Arc<Unbundler>) -> Self {
    StatsServer { config, unbundler }
  }

  pub async fn spawn(
    self,
    notify_sender: Sender<Thread>,
  ) -> Result<(Server, JoinHandle<Result<()>>, ThreadHandle)> {
    let (tx, rx) = mpsc::channel();

    let (join_handle, thread_handle) = thread::spawn(notify_sender, move || {
      let sys = System::new("stats-http-server");

      let unbundler = self.unbundler.clone();
      let server_address = self.config.address;

      let srv = HttpServer::new(move || {
        App::new()
          .wrap(Logger::default())
          .route("/", actix_web::web::get().to(get_status))
          .route("/health", actix_web::web::get().to(get_health))
          .data(State {
            unbundler: Arc::clone(&unbundler),
          })
      })
      .bind(server_address)
      .context(Bind)?
      .shutdown_timeout(60)
      .run();

      tx.send(srv).context(ChannelSend)?;

      sys.run().context(SystemRun)
    });

    Ok((
      rx.recv().context(ChannelReceive)?,
      join_handle,
      thread_handle,
    ))
  }
}

#[derive(Serialize)]
struct GetStatusResponse {
  status: UnbundlerStatus,
  serve_dir: Option<PathBuf>,
}

async fn get_status(data: actix_web::web::Data<State>) -> impl Responder {
  let result: anyhow::Result<String> = try {
    let status = data.unbundler.get_status()?;
    let serve_dir = data.unbundler.get_serve_dir()?;

    serde_json::to_string(&GetStatusResponse { status, serve_dir })?
  };

  match result {
    Ok(response) => HttpResponse::Ok().body(response),
    Err(_) => HttpResponse::InternalServerError().body("Internal Server Error"),
  }
}

async fn get_health(data: actix_web::web::Data<State>) -> impl Responder {
  match data.unbundler.get_status() {
    Ok(status) => match status {
      UnbundlerStatus::Ready => HttpResponse::Ok().body("Ready"),
      UnbundlerStatus::Polling => HttpResponse::Ok().body("Ready"),
      _ => HttpResponse::ServiceUnavailable().body("Not Ready"),
    },
    Err(_) => HttpResponse::InternalServerError().body("Internal Server Error"),
  }
}
