//! Multithreading Utilities.

use mpsc::Sender;
use snafu::Snafu;
use std::{
  sync::{mpsc, Arc, RwLock},
  thread::JoinHandle,
};

#[derive(Snafu, Debug, PartialEq)]
pub enum Error {
  // Unable to obtain a read lock to check the status of a thread.
  LockRead,
}

pub type Result<T, E = Error> = std::result::Result<T, E>;

/// A lightweight abstration over a regular thread that provides an API for
/// determining if a thread has terminated.
pub struct ThreadHandle {
  ended: Arc<RwLock<bool>>,
}

impl ThreadHandle {
  /// Attempts to check if the thread has ended.
  ///
  /// An error may be returned if the underlying channel is disconnected.
  pub fn has_ended(&self) -> Result<bool> {
    let result = self.ended.read().map_err(|_| Error::LockRead)?;

    Ok(*result)
  }
}

/// Like `std::thread::spawn`, but returns a `ThreadHandle` instead.
///
/// # Examples
///
/// Create ten threads and wait for all threads to finish.
///
/// ```
/// use espresso::thread::spawn;
/// use std::{
///    collections::HashMap,
///    sync::{mpsc, Arc, Barrier},
///  };
///
/// let (monitor_tx, monitor_rx) = mpsc::channel();
/// let barrier = Arc::new(Barrier::new(10));
///
/// let mut handles = HashMap::new();
///
/// for _ in 0..10 {
///   let bc = barrier.clone();
///
///   let (join_handle, thread_handle) = spawn(monitor_tx.clone(), move || {
///     /// Sync all threads.
///     bc.wait();
///   });
///
///   handles.insert(join_handle.thread().id(), thread_handle);
/// }
///
/// // Loop until we have been notified of every thread ending.
/// loop {
///   let thread = monitor_rx.recv().unwrap();
///
///   handles.remove(&thread.id());
///
///   if handles.is_empty() {
///     break;
///   }
/// }
/// ```
pub fn spawn<F, T>(
  notify_sender: Sender<std::thread::Thread>,
  f: F,
) -> (JoinHandle<T>, ThreadHandle)
where
  F: FnOnce() -> T,
  F: Send + 'static,
  T: Send + 'static,
{
  let ended = Arc::new(RwLock::new(false));
  let ended_for_spawn = ended.clone();

  let join_handle = std::thread::spawn(move || {
    let ended = ended_for_spawn.clone();

    let result = f();

    let mut ended = ended.write().unwrap();
    *ended = true;
    notify_sender.send(std::thread::current()).unwrap();

    result
  });

  (join_handle, ThreadHandle { ended })
}

#[cfg(test)]
mod tests {
  use super::spawn;
  use std::{
    collections::HashMap,
    sync::{mpsc, Arc, Barrier},
  };

  #[test]
  fn test_spawn() {
    let (monitor_tx, monitor_rx) = mpsc::channel();
    let (ready_tx, ready_rx) = mpsc::channel();
    let (end_tx, end_rx) = mpsc::channel();

    let (join_handle, handle) = spawn(monitor_tx, move || {
      ready_tx.send(()).unwrap();

      end_rx.recv().unwrap();
    });

    ready_rx.recv().unwrap();

    assert_eq!((&handle).has_ended(), Ok(false));

    end_tx.send(()).unwrap();

    monitor_rx.recv().unwrap();
    join_handle.join().unwrap();

    assert_eq!(handle.has_ended(), Ok(true));
  }

  #[test]
  fn test_multiple() {
    let (monitor_tx, monitor_rx) = mpsc::channel();
    let barrier = Arc::new(Barrier::new(11));

    let mut handles = HashMap::new();

    for _ in 0..10 {
      let bc = barrier.clone();

      let (join_handle, thread_handle) = spawn(monitor_tx.clone(), move || {
        bc.wait();
      });

      handles.insert(join_handle.thread().id(), thread_handle);
    }

    for (_, handle) in &handles {
      assert_eq!(handle.has_ended(), Ok(false));
    }

    barrier.wait();

    loop {
      let thread = monitor_rx.recv().unwrap();

      handles.remove(&thread.id());

      if handles.is_empty() {
        break;
      }
    }
  }
}
